@@ -20,7 +20,7 @@ License along with this library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 """
 
-__version__ = "0.4.2"
+__version__ = "0.4.3"
 
 from libxml2dom.macrolib import *
 from libxml2dom.macrolib import \
@@ -30,6 +30,7 @@ from libxml2dom.macrolib import \
     toString as Node_toString, toStream as Node_toStream, \
     toFile as Node_toFile
 import urllib # for parseURI in HTML mode
+import xml.dom # for getElementById
 
 class Implementation(object):
 
@@ -113,6 +114,11 @@ class NamedNodeMap(object):
         self.node.removeAttributeNS(ns, localName)
         return old
 
+    # Iterator emulation.
+
+    def __iter__(self):
+        return NamedNodeMapIterator(self)
+
     # Dictionary emulation methods.
 
     def __getitem__(self, name):
@@ -148,6 +154,22 @@ class NamedNodeMap(object):
 
     length = property(_length)
 
+class NamedNodeMapIterator(object):
+
+    "An iterator over a NamedNodeMap."
+
+    def __init__(self, nodemap):
+        self.nodemap = nodemap
+        self.items = self.nodemap.items()
+
+    def next(self):
+        if self.items:
+            current = self.items[0][1]
+            self.items = self.items[1:]
+            return current
+        else:
+            raise StopIteration
+
 class NodeList(list):
 
     "A wrapper around node lists."
@@ -196,6 +218,12 @@ class Node(object):
 
         return NodeList([self.impl.get_node(_node, self) for _node in Node_childNodes(self._node)])
 
+    def _firstChild(self):
+        return (self.childNodes or [None])[0]
+
+    def _lastChild(self):
+        return (self.childNodes or [None])[-1]
+
     def _attributes(self):
         return NamedNodeMap(self, self.impl)
 
@@ -235,7 +263,11 @@ class Node(object):
         return self.impl.get_node_or_none(Node_nextSibling(self._node), self)
 
     def _doctype(self):
-        return self.impl.get_node(Node_doctype(self._node), self)
+        _doctype = Node_doctype(self._node)
+        if _doctype is not None:
+            return self.impl.get_node(_doctype, self)
+        else:
+            return None
 
     def _publicId(self):
         # NOTE: To be fixed when the libxml2mod API has been figured out.
@@ -340,18 +372,28 @@ class Node(object):
         return self.importNode(self, deep)
 
     def insertBefore(self, tmp, oldNode):
+        if tmp.ownerDocument != self.ownerDocument:
+            raise xml.dom.DOMException(xml.dom.WRONG_DOCUMENT_ERR)
+        if oldNode.parentNode != self:
+            raise xml.dom.DOMException(xml.dom.NOT_FOUND_ERR)
         if hasattr(tmp, "as_native_node"):
             return self.impl.get_node(Node_insertBefore(self._node, tmp.as_native_node(), oldNode.as_native_node()), self)
         else:
             return self.impl.get_node(Node_insertBefore(self._node, tmp, oldNode.as_native_node()), self)
 
     def replaceChild(self, tmp, oldNode):
+        if tmp.ownerDocument != self.ownerDocument:
+            raise xml.dom.DOMException(xml.dom.WRONG_DOCUMENT_ERR)
+        if oldNode.parentNode != self:
+            raise xml.dom.DOMException(xml.dom.NOT_FOUND_ERR)
         if hasattr(tmp, "as_native_node"):
             return self.impl.get_node(Node_replaceChild(self._node, tmp.as_native_node(), oldNode.as_native_node()), self)
         else:
             return self.impl.get_node(Node_replaceChild(self._node, tmp, oldNode.as_native_node()), self)
 
     def appendChild(self, tmp):
+        if tmp.ownerDocument != self.ownerDocument:
+            raise xml.dom.DOMException(xml.dom.WRONG_DOCUMENT_ERR)
         if hasattr(tmp, "as_native_node"):
             return self.impl.get_node(Node_appendChild(self._node, tmp.as_native_node()), self)
         else:
@@ -362,6 +404,15 @@ class Node(object):
             Node_removeChild(self._node, tmp.as_native_node())
         else:
             Node_removeChild(self._node, tmp)
+        return tmp
+
+    def getElementById(self, identifier):
+        nodes = self.xpath(".//*[@xml:id='" + identifier.replace("'", "&apos;") + "']",
+            namespaces={"xml" : xml.dom.XML_NAMESPACE})
+        if nodes:
+            return nodes[0]
+        else:
+            return None
 
     def getElementsByTagName(self, tagName):
         return self.xpath(".//" + tagName)
@@ -389,6 +440,8 @@ class Node(object):
         self.replaceChild(self.ownerDocument.createTextNode("".join(texts)), text_nodes[-1])
 
     childNodes = property(_childNodes)
+    firstChild = property(_firstChild)
+    lastChild = property(_lastChild)
     value = data = nodeValue = property(_nodeValue, _setNodeValue)
     textContent = property(_textContent)
     name = nodeName = property(_nodeName)
@@ -413,8 +466,11 @@ class Node(object):
     def isSameNode(self, other):
         return self == other
 
+    def __hash__(self):
+        return hash(self.localName)
+
     def __eq__(self, other):
-        return isinstance(other, Node) and libxml2mod.xmlXPathCmpNodes(self._node, other._node) == 0
+        return isinstance(other, Node) and Node_equals(self._node, other._node)
 
     def __ne__(self, other):
         return not (self == other)
@@ -522,15 +578,22 @@ def createDocumentType(localName, publicId, systemId):
 def createDocument(namespaceURI, localName, doctype):
     return default_impl.createDocument(namespaceURI, localName, doctype)
 
-def parse(stream_or_string, html=0, htmlencoding=None, impl=None):
+def parse(stream_or_string, html=0, htmlencoding=None, unfinished=0, impl=None):
 
     """
     Parse the given 'stream_or_string', where the supplied object can either be
     a stream (such as a file or stream object), or a string (containing the
-    filename of a document). If the optional 'html' parameter is set to a true
-    value, the content to be parsed will be treated as being HTML rather than
-    XML. If the optional 'htmlencoding' is specified, HTML parsing will be
-    performed with the document encoding assumed to that specified.
+    filename of a document). The optional parameters described below should be
+    provided as keyword arguments.
+
+    If the optional 'html' parameter is set to a true value, the content to be
+    parsed will be treated as being HTML rather than XML. If the optional
+    'htmlencoding' is specified, HTML parsing will be performed with the
+    document encoding assumed to that specified.
+
+    If the optional 'unfinished' parameter is set to a true value, unfinished
+    documents will be parsed, even though such documents may be missing content
+    such as closing tags.
 
     A document object is returned by this function.
     """
@@ -539,48 +602,66 @@ def parse(stream_or_string, html=0, htmlencoding=None, impl=None):
 
     if hasattr(stream_or_string, "read"):
         stream = stream_or_string
-        return parseString(stream.read(), html, htmlencoding, impl)
+        return parseString(stream.read(), html=html, htmlencoding=htmlencoding, unfinished=unfinished, impl=impl)
     else:
-        return parseFile(stream_or_string, html, htmlencoding, impl)
+        return parseFile(stream_or_string, html=html, htmlencoding=htmlencoding, unfinished=unfinished, impl=impl)
 
-def parseFile(filename, html=0, htmlencoding=None, impl=None):
+def parseFile(filename, html=0, htmlencoding=None, unfinished=0, impl=None):
 
     """
-    Parse the file having the given 'filename'. If the optional 'html' parameter
-    is set to a true value, the content to be parsed will be treated as being
-    HTML rather than XML. If the optional 'htmlencoding' is specified, HTML
-    parsing will be performed with the document encoding assumed to be that
-    specified.
+    Parse the file having the given 'filename'. The optional parameters
+    described below should be provided as keyword arguments.
+
+    If the optional 'html' parameter is set to a true value, the content to be
+    parsed will be treated as being HTML rather than XML. If the optional
+    'htmlencoding' is specified, HTML parsing will be performed with the
+    document encoding assumed to that specified.
+
+    If the optional 'unfinished' parameter is set to a true value, unfinished
+    documents will be parsed, even though such documents may be missing content
+    such as closing tags.
 
     A document object is returned by this function.
     """
 
     impl = impl or default_impl
-    return impl.adoptDocument(Node_parseFile(filename, html, htmlencoding))
+    return impl.adoptDocument(Node_parseFile(filename, html=html, htmlencoding=htmlencoding, unfinished=unfinished))
 
-def parseString(s, html=0, htmlencoding=None, impl=None):
+def parseString(s, html=0, htmlencoding=None, unfinished=0, impl=None):
 
     """
-    Parse the content of the given string 's'. If the optional 'html' parameter
-    is set to a true value, the content to be parsed will be treated as being
-    HTML rather than XML. If the optional 'htmlencoding' is specified, HTML
-    parsing will be performed with the document encoding assumed to be that
-    specified.
+    Parse the content of the given string 's'. The optional parameters described
+    below should be provided as keyword arguments.
+
+    If the optional 'html' parameter is set to a true value, the content to be
+    parsed will be treated as being HTML rather than XML. If the optional
+    'htmlencoding' is specified, HTML parsing will be performed with the
+    document encoding assumed to that specified.
+
+    If the optional 'unfinished' parameter is set to a true value, unfinished
+    documents will be parsed, even though such documents may be missing content
+    such as closing tags.
 
     A document object is returned by this function.
     """
 
     impl = impl or default_impl
-    return impl.adoptDocument(Node_parseString(s, html, htmlencoding))
+    return impl.adoptDocument(Node_parseString(s, html=html, htmlencoding=htmlencoding, unfinished=unfinished))
 
-def parseURI(uri, html=0, htmlencoding=None, impl=None):
+def parseURI(uri, html=0, htmlencoding=None, unfinished=0, impl=None):
 
     """
-    Parse the content found at the given 'uri'. If the optional 'html' parameter
-    is set to a true value, the content to be parsed will be treated as being
-    HTML rather than XML. If the optional 'htmlencoding' is specified, HTML
-    parsing will be performed with the document encoding assumed to be that
-    specified.
+    Parse the content found at the given 'uri'. The optional parameters
+    described below should be provided as keyword arguments.
+
+    If the optional 'html' parameter is set to a true value, the content to be
+    parsed will be treated as being HTML rather than XML. If the optional
+    'htmlencoding' is specified, HTML parsing will be performed with the
+    document encoding assumed to that specified.
+
+    If the optional 'unfinished' parameter is set to a true value, unfinished
+    documents will be parsed, even though such documents may be missing content
+    such as closing tags.
 
     XML documents are retrieved using libxml2's own network capabilities; HTML
     documents are retrieved using the urllib module provided by Python. To
@@ -599,12 +680,12 @@ def parseURI(uri, html=0, htmlencoding=None, impl=None):
     if html:
         f = urllib.urlopen(uri)
         try:
-            return parse(f, html, htmlencoding, impl)
+            return parse(f, html=html, htmlencoding=htmlencoding, unfinished=unfinished, impl=impl)
         finally:
             f.close()
     else:
         impl = impl or default_impl
-        return impl.adoptDocument(Node_parseURI(uri, html, htmlencoding))
+        return impl.adoptDocument(Node_parseURI(uri, html=html, htmlencoding=htmlencoding, unfinished=unfinished))
 
 def toString(node, encoding=None, prettyprint=0):
 
