@@ -49,10 +49,6 @@ def to_unicode(s):
     else:
         return s
 
-def check_document(node, tmp):
-    if Node_ownerDocument(tmp) is not None and Node_ownerDocument(node) is not Node_ownerDocument(tmp):
-        raise xml.dom.DOMException(xml.dom.WRONG_DOCUMENT_ERR)
-
 def get_ns(ns):
     out_ns = to_unicode(libxml2mod.xmlNodeGetContent(ns))
     # Detect "" and produce None as the empty namespace.
@@ -162,6 +158,9 @@ _reverseNodeTypes = {}
 for label, value in _nodeTypes.items():
     _reverseNodeTypes[value] = label
 
+def Node_equals(node, other):
+    return libxml2mod.xmlXPathCmpNodes(node, other) == 0
+
 def Node_ownerDocument(node):
     return libxml2mod.doc(node)
 
@@ -417,15 +416,12 @@ def Node_createCDATASection(node, value):
     return libxml2mod.xmlNewCDataBlock(Node_ownerDocument(node), value, len(value))
 
 def Node_insertBefore(node, tmp, oldNode):
-    check_document(node, tmp)
     return libxml2mod.xmlAddPrevSibling(oldNode, tmp)
 
 def Node_replaceChild(node, tmp, oldNode):
-    check_document(node, tmp)
     return libxml2mod.xmlReplaceNode(oldNode, tmp)
 
 def Node_appendChild(node, tmp):
-    check_document(node, tmp)
     return libxml2mod.xmlAddChild(node, tmp)
 
 def Node_removeChild(node, child):
@@ -499,6 +495,15 @@ def Node_xpath(node, expr, variables=None, namespaces=None):
     libxml2mod.xmlXPathFreeContext(context)
     return result
 
+# Exceptions.
+
+class LSException(Exception):
+
+    "DOM Level 3 Load/Save exception."
+
+    PARSE_ERR = 81
+    SERIALIZE_ERR = 82
+
 # Utility functions.
 
 def createDocument(namespaceURI, localName, doctype):
@@ -512,49 +517,55 @@ def createDocument(namespaceURI, localName, doctype):
         libxml2mod.xmlCreateIntSubset(d, doctype.localName, doctype.publicId, doctype.systemId)
     return d
 
-def parse(stream_or_string, html=0, htmlencoding=None):
+def parse(stream_or_string, html=0, htmlencoding=None, unfinished=0):
     if hasattr(stream_or_string, "read"):
         stream = stream_or_string
-        return parseString(stream.read(), html, htmlencoding)
+        return parseString(stream.read(), html=html, htmlencoding=htmlencoding, unfinished=unfinished)
     else:
-        return parseFile(stream_or_string, html, htmlencoding)
+        return parseFile(stream_or_string, html=html, htmlencoding=htmlencoding, unfinished=unfinished)
 
-def parseFile(s, html=0, htmlencoding=None):
+def parseFile(s, html=0, htmlencoding=None, unfinished=0):
     # NOTE: Switching off validation and remote DTD resolution.
     if not html:
         context = libxml2mod.xmlCreateFileParserCtxt(s)
-        libxml2mod.xmlParserSetPedantic(context, 0)
-        libxml2mod.xmlParserSetValidate(context, 0)
-        libxml2mod.xmlCtxtUseOptions(context, XML_PARSE_NOERROR | XML_PARSE_NOWARNING | XML_PARSE_NONET)
-        libxml2mod.xmlParseDocument(context)
-        return libxml2mod.xmlParserGetDoc(context)
+        Parser_configure(context)
+        Parser_parse(context)
+        doc = Parser_document(context)
+        if unfinished or Parser_well_formed(context):
+            return doc
+        else:
+            raise LSException(LSException.PARSE_ERR)
     else:
         return libxml2mod.htmlReadFile(s, htmlencoding, HTML_PARSE_NOERROR | HTML_PARSE_NOWARNING | HTML_PARSE_NONET)
 
-def parseString(s, html=0, htmlencoding=None):
+def parseString(s, html=0, htmlencoding=None, unfinished=0):
     # NOTE: Switching off validation and remote DTD resolution.
     if not html:
         context = libxml2mod.xmlCreateMemoryParserCtxt(s, len(s))
-        libxml2mod.xmlParserSetPedantic(context, 0)
-        libxml2mod.xmlParserSetValidate(context, 0)
-        libxml2mod.xmlCtxtUseOptions(context, XML_PARSE_NOERROR | XML_PARSE_NOWARNING | XML_PARSE_NONET)
-        libxml2mod.xmlParseDocument(context)
-        return libxml2mod.xmlParserGetDoc(context)
+        Parser_configure(context)
+        Parser_parse(context)
+        doc = Parser_document(context)
+        if unfinished or Parser_well_formed(context):
+            return doc
+        else:
+            raise LSException(LSException.PARSE_ERR)
     else:
         # NOTE: URL given as None.
         html_url = None
         return libxml2mod.htmlReadMemory(s, len(s), html_url, htmlencoding,
             HTML_PARSE_NOERROR | HTML_PARSE_NOWARNING | HTML_PARSE_NONET)
 
-def parseURI(uri, html=0, htmlencoding=None):
+def parseURI(uri, html=0, htmlencoding=None, unfinished=0):
     # NOTE: Switching off validation and remote DTD resolution.
     if not html:
         context = libxml2mod.xmlCreateURLParserCtxt(uri, 0)
-        libxml2mod.xmlParserSetPedantic(context, 0)
-        libxml2mod.xmlParserSetValidate(context, 0)
-        libxml2mod.xmlCtxtUseOptions(context, XML_PARSE_NOERROR | XML_PARSE_NOWARNING | XML_PARSE_NONET)
-        libxml2mod.xmlParseDocument(context)
-        return libxml2mod.xmlParserGetDoc(context)
+        Parser_configure(context)
+        Parser_parse(context)
+        doc = Parser_document(context)
+        if unfinished or Parser_well_formed(context):
+            return doc
+        else:
+            raise LSException(LSException.PARSE_ERR)
     else:
         raise NotImplementedError, "parseURI does not yet support HTML"
 
@@ -567,7 +578,7 @@ def toStream(node, stream, encoding=None, prettyprint=0):
 def toFile(node, f, encoding=None, prettyprint=0):
     libxml2mod.saveNodeTo(node, f, encoding, prettyprint)
 
-# libxml2mod constants.
+# libxml2mod constants and helper functions.
 
 HTML_PARSE_NOERROR = 32
 HTML_PARSE_NOWARNING = 64
@@ -576,4 +587,24 @@ XML_PARSE_NOERROR = 32
 XML_PARSE_NOWARNING = 64
 XML_PARSE_NONET = 2048
 
+def Parser_push():
+    return libxml2mod.xmlCreatePushParser(None, "", 0, None)
+
+def Parser_configure(context):
+    libxml2mod.xmlParserSetPedantic(context, 0)
+    libxml2mod.xmlParserSetValidate(context, 0)
+    libxml2mod.xmlCtxtUseOptions(context, XML_PARSE_NOERROR | XML_PARSE_NOWARNING | XML_PARSE_NONET)
+
+def Parser_feed(context, s):
+    libxml2mod.xmlParseChunk(context, s, len(s), 1)
+
+def Parser_well_formed(context):
+    return libxml2mod.xmlParserGetWellFormed(context)
+
+def Parser_document(context):
+    return libxml2mod.xmlParserGetDoc(context)
+
+def Parser_parse(context):
+    libxml2mod.xmlParseDocument(context)
+
 # vim: tabstop=4 expandtab shiftwidth=4
