@@ -28,6 +28,7 @@ from libxml2dom.macrolib import *
 from libxml2dom.macrolib import \
     createDocument as Node_createDocument
 import xml.dom
+import urllib
 import math
 import re
 
@@ -80,6 +81,9 @@ class SVGImplementation(libxml2dom.Implementation):
         else:
             return libxml2dom.Implementation.get_node(self, _node, context_node)
 
+    def get_global(self, doc):
+        return SVGGlobal(doc)
+
     # Convenience functions.
 
     def createSVGDocument(self):
@@ -141,7 +145,23 @@ class ElementTraversal:
     nextElementSibling = property(_nextElementSibling)
     previousElementSibling = property(_previousElementSibling)
 
-class SVGGlobal: # Global, EventListenerInitializer2
+class EventListenerInitializer2:
+
+    "An event listener initialisation interface."
+
+    def initializeEventListeners(self, scriptElement):
+        pass
+
+    def createEventListener(self, handlerElement):
+        pass
+
+class Global:
+
+    "An empty global interface."
+
+    pass
+
+class SVGGlobal(Global, EventListenerInitializer2):
 
     "An SVG global."
 
@@ -151,6 +171,10 @@ class SVGGlobal: # Global, EventListenerInitializer2
 
         self.document = document
 
+        # Listener management.
+
+        self.listeners = {}
+
     def createConnection(self):
         raise NotImplementedError, "createConnection"
 
@@ -161,10 +185,16 @@ class SVGGlobal: # Global, EventListenerInitializer2
         raise NotImplementedError, "gotoLocation"
 
     def binaryToString(self, octets, encoding):
-        raise NotImplementedError, "binaryToString"
+        try:
+            return unicode(octets, encoding)
+        except UnicodeDecodeError, exc:
+            raise GlobalException(GlobalException.ENCODING_ERR)
 
     def stringToBinary(self, data, encoding):
-        raise NotImplementedError, "stringToBinary"
+        try:
+            return data.encode(encoding)
+        except UnicodeEncodeError, exc:
+            raise GlobalException(GlobalException.ENCODING_ERR)
 
     def getURL(self, iri, callback):
 
@@ -183,8 +213,27 @@ class SVGGlobal: # Global, EventListenerInitializer2
         finally:
             f.close()
 
-    def postURL(self, iri, data, callback, type, encoding):
-        raise NotImplementedError, "postURL"
+    def postURL(self, iri, data, callback, type=None, encoding=None):
+
+        # NOTE: Not asynchronous.
+        # NOTE: The urlopen function may not support IRIs.
+        # No exceptions are supposed to be raised, which is a bit nasty.
+
+        opener = urllib.URLopener()
+        opener.addheader("Content-Type", type or "text/plain")
+        if encoding:
+            opener.addheader("Content-Encoding", encoding)
+        f = opener.open(iri, data)
+        try:
+            try:
+                content = f.read()
+                contentType = f.headers["Content-Type"]
+                callback.operationComplete(AsyncURLStatus(1, contentType, content))
+            except:
+                callback.operationComplete(AsyncURLStatus(0, None, None))
+        finally:
+            f.close()
+            opener.close()
 
     def parseXML(self, data, contextDoc):
         doc = parseString(data)
@@ -641,7 +690,10 @@ class SVGNode(libxml2dom.Node):
             namespaces["svg"] = SVG_NAMESPACE
         return libxml2dom.Node.xpath(self, expr, variables, namespaces)
 
-class SVGDocument(libxml2dom._Document, SVGNode, DocumentEvent, EventTarget):
+# NOTE: DocumentEvent is from DOM Level 3 Events.
+# NOTE: EventSystem is a special libxml2dom.events class.
+
+class SVGDocument(libxml2dom._Document, SVGNode, EventTarget, DocumentEvent, EventSystem):
 
     "An SVG-specific document node."
 
@@ -653,12 +705,15 @@ class SVGDocument(libxml2dom._Document, SVGNode, DocumentEvent, EventTarget):
         """
 
         libxml2dom._Document.__init__(self, node, impl)
-        self.global_ = SVGGlobal(self) # parent
+        self.global_ = self.impl.get_global(self) # parent
 
 class SVGElement(SVGNode, EventTarget, TraitAccess, ElementTraversal): # NOTE: SVGNode instead of Element.
 
     "An SVG-specific element."
 
+    def __init__(self, *args, **kw):
+        SVGNode.__init__(self, *args, **kw)
+
     def _id(self):
         return self.getAttribute("id")
 
@@ -763,6 +818,37 @@ class SVGSVGElement(SVGLocatableElement, SVGTimedElement):
     currentTranslate = property(_currentTranslate)
     viewport = property(_viewport)
 
+# Event handler initialisation.
+
+def initialiseEvents(doc):
+
+    """
+    See: http://www.w3.org/TR/SVGMobile12/svgudom.html#svg__EventListenerInitializer2
+    See: http://www.w3.org/TR/xml-events/#section-listener-element
+    """
+
+    # Initialise script element listeners.
+
+    for script in doc.xpath("//svg:script"):
+        doc.global_.initializeEventListeners(script)
+
+    # Initialise handler element listeners using XML Events.
+
+    for handler in doc.xpath("//svg:handler"):
+        listener = doc.global_.createEventListener(handler)
+
+        # Attempt to parameterise the registration using the XML Events attributes.
+
+        phase = handler.getAttributeNS(libxml2dom.events.XML_EVENTS_NAMESPACE, "phase")
+
+        # Add the listener for the appropriate type and phases.
+
+        handler.parentNode.addEventListener(
+            handler.getAttributeNS(libxml2dom.events.XML_EVENTS_NAMESPACE, "event"),
+            listener,
+            phase == "capture"
+            )
+
 # Utility functions.
 
 createDocument = libxml2dom.createDocument
@@ -771,17 +857,25 @@ createDocumentType = libxml2dom.createDocumentType
 def createSVGDocument():
     return default_impl.createSVGDocument()
 
-def parse(stream_or_string, html=0, htmlencoding=None):
-    return libxml2dom.parse(stream_or_string, html, htmlencoding, default_impl)
+def parse(stream_or_string, html=0, htmlencoding=None, unfinished=0, impl=None):
+    doc = libxml2dom.parse(stream_or_string, html=html, htmlencoding=htmlencoding, unfinished=unfinished, impl=(impl or default_impl))
+    initialiseEvents(doc)
+    return doc
 
-def parseFile(filename, html=0, htmlencoding=None):
-    return libxml2dom.parseFile(filename, html, htmlencoding, default_impl)
+def parseFile(filename, html=0, htmlencoding=None, unfinished=0, impl=None):
+    doc = libxml2dom.parseFile(filename, html=html, htmlencoding=htmlencoding, unfinished=unfinished, impl=(impl or default_impl))
+    initialiseEvents(doc)
+    return doc
 
-def parseString(s, html=0, htmlencoding=None):
-    return libxml2dom.parseString(s, html, htmlencoding, default_impl)
+def parseString(s, html=0, htmlencoding=None, unfinished=0, impl=None):
+    doc = libxml2dom.parseString(s, html=html, htmlencoding=htmlencoding, unfinished=unfinished, impl=(impl or default_impl))
+    initialiseEvents(doc)
+    return doc
 
-def parseURI(uri, html=0, htmlencoding=None):
-    return libxml2dom.parseURI(uri, html, htmlencoding, default_impl)
+def parseURI(uri, html=0, htmlencoding=None, unfinished=0, impl=None):
+    doc = libxml2dom.parseURI(uri, html=html, htmlencoding=htmlencoding, unfinished=unfinished, impl=(impl or default_impl))
+    initialiseEvents(doc)
+    return doc
 
 # Single instance of the implementation.
 
